// header file of picoexif.c

// number of directory items
#define EXIF_MAXITEMS 16

// amount of memory for string variables (and later other arrays)
#define EXIF_MAXVARBUFLEN 128




// allow printing eror messages to stderr, for debug
#define EXIF_DEBUG_PRINTERR

// zero exif_buffer before use
//#define EXIF_WIPEBUFFER
// fill exif_buffer with '@' before use, to see "leaks"
#define EXIF_DEBUG_MARKBUFFER



// offsets, dictated by EXIF/JFIF specs
#define EXIF_INTERNAL_ITEMLEN 12           // length of directory record
#define EXIF_INTERNAL_BUF_FIRSTDIR 0x14    // offset of the first directory record
#define EXIF_INTERNAL_BUF_ITEMSOFFS 0x12   // position for number of items, 2-byte int, endianness
#define EXIF_INTERNAL_DIR_PADDING 4        // the zero bytes after the directory end

// sizes and offsets calculated from maxitems and maxvarbuflen
#define EXIF_BUF_VARBUFSTART (EXIF_INTERNAL_BUF_FIRSTDIR + (EXIF_MAXITEMS*EXIF_INTERNAL_ITEMLEN) + EXIF_INTERNAL_DIR_PADDING ) // end of header/dir item portion, start of variable-length var buffer
#define EXIF_MAXBUF (EXIF_INTERNAL_BUF_FIRSTDIR + (EXIF_MAXITEMS*EXIF_INTERNAL_ITEMLEN) + EXIF_MAXVARBUFLEN + EXIF_INTERNAL_DIR_PADDING +2 )  // 2 bytes to guard against off-by-one errors




//  EXIF file format: https://www.media.mit.edu/pia/Research/deepview/exif.html
//  EXIF tag type list: https://www.exiv2.org/tags.html

// offsets in the 12-byte record
#define EXIF_RECORDOFFSET_ITEM 0
#define EXIF_RECORDOFFSET_TYPE 2
#define EXIF_RECORDOFFSET_LEN 4
#define EXIF_RECORDOFFSET_VAL 8

// EXIF variable types, from specs
// vartype type   length
//  0      --
//  1      uint8  1
//  2      string 1   items=number of bytes
//  3      uint16 2
//  4      uint32 4
//  5      uratio 4   2 bytes for numerator, then 2 bytes for denominator
//  6      sint8  1
//  7      undef  1
//  8      sint16 2
//  9      sint32 4
// 10      sratio 8   WARN: won't fit 4-byte space, not supported yet
// 11      float  4
// 12      double 8   WARN: won't fit 4-byte space, not supported yet

#define EXIF_VARTYPE_UINT8    1
#define EXIF_VARTYPE_STR      2
#define EXIF_VARTYPE_UINT16   3
#define EXIF_VARTYPE_UINT32   4
#define EXIF_VARTYPE_URATIO   5
#define EXIF_VARTYPE_SINT8    6
#define EXIF_VARTYPE_UNDEF    7
#define EXIF_VARTYPE_SINT16   8
#define EXIF_VARTYPE_SINT32   9
#define EXIF_VARTYPE_SRATIO  10
#define EXIF_VARTYPE_FLOAT   11
#define EXIF_VARTYPE_DOUBLE  12

#define EXIF_VARTYPE_ISPOINTER 0x1000


// errors
#define EXIF_ERR_NONE 0
#define EXIF_ERR_TOOMANYITEMS 1
#define EXIF_ERR_TOOMANYITEMS_STR 2
#define EXIF_ERR_STRTOOLONG 3



// buffer itself
extern unsigned char exif_buf[];  // raw buffer for data

// counters and pointers
extern int exif_itemcnt; // incremented by exif_additem
extern int exif_varbuflen; // offset from start of varbuffer
extern int exif_errno; // did error happen?

// user function, main API call for adding a 1..4 byte long variable to directory array
// value: for 1..4 bytes the value itself; otherwise needs offset to buffer, which is implemented only for exif_additemstr()
//   type is the EXIF record number
//   vartype is variable type, EXIF_VARTYPE_*
//   items is number of items, usually 1 for numbers, 1..4 for bytes/characters
//   val is the value of the record
// ASSUMPTION: only up to 4-byte data here!!!
// CAUTION: no support for larger arrays, optional var-buffer allocation is via exif_addbufstr, and pointer goes here
void exif_additem(int type,int vartype,long items,long val);


// user function, main API call for adding a string variable to directory array and to var-buffer
//   type is the EXIF record number
//   s is the string value, variable of EXIF_VARTYPE_STR
void exif_additemstr(int type,char*s);


// user function, call to initialize the array; prepares the structure
// call as the first thing
void exif_init();

// user function, returns total header length; to know the output length
int exif_gethdrlen();

// user function, call to finish the exif generation; calculates lengths and adjusts pointers
// call after adding all the records
void exif_finish();


// output:
//
// header and fixed-length directory items
//  for(int t=0;t<exif_gethdrlen();t++)
//    putchar(exif_buf[t]);
//
// variable-size buffer
//  for(int t=0;t<exif_varbuflen;t++)
//    putchar(exif_buf[EXIF_BUF_VARBUFSTART+t]);




/*
// demo call for setting exif variables
void demo_setexif(){
  exif_init();

  exif_additemstr(0x0110,"TEST2-12345678");                   //  digicam model
  //exif_additem(0x0112,EXIF_VARTYPE_UINT16,  1,1);                 //  orientation; 1=upper left, 3=lower right, 6=upper right, 8=lower left, 9=undefined
  exif_additem(0x9000,EXIF_VARTYPE_UNDEF,   4,0x30313230);        //  EXIF version; 4-chars, "0120"
  exif_additemstr(0x9003,"2018:09:06 18:18:00");              //  time created
  exif_additemstr(0x9291,"0123");                             //  time created, fraction of second
  exif_additemstr(0x9004,"2019:09:06 19:19:00");              //  time edited
  exif_additem(0x9999,EXIF_VARTYPE_UNDEF,   4,0x41414141);        //  undefined tag, for test
  //exif_additemstr(0x9291,"012346");                             //  time created, fraction of second

  exif_finish();
}



void demo_dumpbin(){
  for(int t=0;t<exif_gethdrlen();t++)putchar(exif_buf[t]);                    // output the fixed-format part
  for(int t=0;t<exif_varbuflen;    t++)putchar(exif_buf[EXIF_BUF_VARBUFSTART+t]);  // output the rest
  //
  fprintf(stderr,"bufoffs=%d\n",exif_varbuflen);
  fprintf(stderr,"MAXBUF=%d\n",EXIF_MAXBUF);
}


int main(){
  demo_setexif();
  demo_dumpbin();
  return 0;
}
*/
